@@ -14,12 +14,20 @@ static void R_abortHandler() {
   error("UCSC library operation failed");
 }
 
+extern int R_ignore_SIGPIPE;
+
 void pushRHandlers() {
   pushAbortHandler(R_abortHandler);
-  pushWarnHandler(R_warnHandler);  
+  pushWarnHandler(R_warnHandler);
+#ifndef WIN32
+  R_ignore_SIGPIPE = 1;
+#endif
 }
 
 void popRHandlers() {
   popAbortHandler();
   popWarnHandler();
+#ifndef WIN32
+  R_ignore_SIGPIPE = 0;
+#endif
 }
